'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' Game15, a simple "15 tiles" brick game, written in PB/DLL 6.0,
' January 1999 by Borje Hagsten. Released as Public Domain.
' Free to enhance, learn from or simply to enjoy and play.
'
' The "rules" are simple. Click on the tiles next to the open
' space in order to move them and try to line them up, 1 to 15.
' A click and blue caption on a brick indicates a correct position.
' It's not as easy as it may look like from the beginning and some
' people may even become addicted to playing it. A complete 32-bit
' game for Windows with a file size of just 20 KB - PB rules!
' Copy, paste, compile with PB/DLL 6.0 and have fun.. :-)
'
' - Useful code in this sample:
' How to build a range of controls in a loop
' How to build "button-like" labels, with a raised edge
' How to change the color of a label's caption text
' How to paint a form's background
' How to detect and move controls at runtime
' How to generate a serie of unique random numbers
' How to use the timer, etc. etc.
'
' - Some ideas, in case you want to do something more of it:
' Use a split bitmap instead of numbers on the labels (StrechBlt)
' Create and add resource file with version info and program icon
' Create a routine to save and read results
' Create an "About" dialog
' Add a menu
' Add sound support (Semen's speaker sound routines maybe..)
' Add help and some game info
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#COMPILE EXE
#DIM ALL
#INCLUDE "WIN32API.INC"

%ID_CLOSE  = 10
%ID_MIX    = 11
%ID_LABEL1 = 20
%ID_LABEL2 = 21
%ID_LABEL3 = 22
%ID_TIMER  = 30
%ID_ONTOP  = 40
%ID_NONTOP = 41

TYPE bx
  x AS LONG ' x pos
  y AS LONG ' y pos
  z AS LONG ' reference number
  c AS LONG ' color value
END TYPE

GLOBAL hDlg     AS LONG
GLOBAL DoPaint  AS LONG
GLOBAL tTimer   AS LONG
GLOBAL brc      AS LONG
GLOBAL n        AS LONG
GLOBAL EmptyPos AS LONG
GLOBAL hTimer   AS LONG
GLOBAL kPos     AS LONG
GLOBAL bPos()   AS bx

DECLARE CALLBACK FUNCTION DlgCallback()
DECLARE FUNCTION EndProcedure() AS LONG
DECLARE FUNCTION CheckDone AS LONG
DECLARE SUB NewGame
DECLARE SUB ReMix

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' PBMAIN, where all controls and the dialog is created
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
FUNCTION PBMAIN
  LOCAL x AS LONG, y AS LONG, z AS LONG, w AS LONG, h AS LONG, b AS LONG
  LOCAL a AS LONG
  brc = 4        'number of columns/rows - use at least 4..
' Note: The brc value is all you have to change. The dialog
' and all controls will resize themselves accordingly. Try
' with 5 or 6 to create a game with 25 or 36 bricks ..  :-)

  w = 26          'width
  h = 24          'height
  a = 10          'distance from dialog edge
  n = brc^2 - 1   'number of bricks
  b = brc - 1     'zero based number of columns and rows
  EmptyPos = n    'initial empty space, down in the corner
  REDIM bPos (n)  'array for storing Brick info

  DIALOG NEW 0, FORMAT$(n) & " tiles",,, brc * w + 22, brc * h + 90, _
                             %WS_SYSMENU OR %WS_CAPTION, 0 TO hDlg

  CONTROL ADD LABEL,  hDlg, %ID_LABEL2, "",       9, 9, _
                      brc * w + 2, brc * h + 2, %SS_SUNKEN
  CONTROL ADD LABEL,  hDlg, %ID_LABEL1, "",       a, brc * h + 18, _
                      brc * w + 2, 12, %SS_CENTER  OR %SS_SUNKEN
  CONTROL ADD LABEL,  hDlg, %ID_LABEL3, "",       a, brc * h + 30, _
                      brc * w + 2, 12, %SS_CENTER  OR %SS_SUNKEN
  CONTROL ADD BUTTON, hDlg, %ID_MIX,    "&Mix",   a, brc * h + 50, brc * w + 2, 14
  CONTROL ADD BUTTON, hDlg, %ID_CLOSE,  "E&xit",  a, brc * h + 68, brc * w + 2, 14

  'A zero based index would be better here, but some low values are already
  'used by the DDT dialog (like IDCANCEL, etc), so we better use another
  'approach and increase it to a "100 based" line of controls instead
  z = 100
  FOR y = 0 TO b
    FOR x = 0 TO b
      bPos(z - 100).x = a + x * w
      bPos(z - 100).y = a + y * h
      CONTROL ADD LABEL, hDlg, z, FORMAT$(z - 99), a + x * w, a + y * h, w, h, _
                         %SS_NOTIFY OR %SS_CENTER OR %WS_BORDER, %BDR_RAISED
      INCR z
      IF z > 99 + n THEN EXIT FOR
    NEXT
  NEXT
  bPos(n).x = a + b * w  'x and y for the inital empty space
  bPos(n).y = a + b * h

  'Create a timer event, just to set the initial table and textcolors for bricks
  'Tried with other solutions, but the blue color for correctly placed bricks
  'simply would not show up. The timer trix works fine though..  :-)
  hTimer = SetTimer(hDlg, %ID_TIMER, 1, BYVAL %NULL)

  DIALOG SHOW MODAL hDlg CALL DlgCallback
END FUNCTION

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' One dialog Callback function takes care of almost everything
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
CALLBACK FUNCTION DlgCallback()
  LOCAL  rStr         AS STRING
  LOCAL  hBrushColor  AS LONG
  STATIC hBrushBlack  AS LONG
  STATIC hBrushBlue   AS LONG
  STATIC hBrushDialog AS LONG

  SELECT CASE CBMSG
    CASE %WM_TIMER
      IF CBCTL = %ID_TIMER THEN
        KillTimer hDlg, hTimer     'Kill first, initial timer
        CALL CheckDone             'Check the bricks and create new timer
        hTimer = SetTimer(hDlg, 77, 1000, BYVAL %NULL) '1000 mSec = 1 sec
      ELSE
        CONTROL SET TEXT hDlg, %ID_LABEL3, _
                    FORMAT$((TIMER - tTimer)  \  60, "0") & " min. " _
                  & FORMAT$(CLNG(TIMER - tTimer) MOD 60, "0") & " sec."
      END IF

    CASE %WM_INITDIALOG 'Create brushs for the label text colors
      LOCAL Lb AS LOGBRUSH
      Lb.lbStyle  = %WHITE_BRUSH

      Lb.lbColor  = %BLACK
      hBrushBlack = CreateBrushIndirect(Lb)
      Lb.lbColor  = %BLUE
      hBrushBlue  = CreateBrushIndirect(Lb)

                     'Create a background brush for the dialog
      Lb.lbColor   = RGB(192,192,156) 'The dialog's color
      hBrushDialog = CreateBrushIndirect(Lb)
      CALL ReMix    'Mix and then let the timer take care of colors

    CASE %WM_COMMAND
      SELECT CASE CBCTL
        CASE %ID_MIX                             'Start a new game..
          rStr = "New game?"
          IF MessageBox(hDlg, BYVAL STRPTR(rStr),"Game" & FORMAT$(n) _
                & CHR$(0), %MB_YESNO OR %MB_ICONQUESTION) = %IDYES THEN
            CALL NewGame
          END IF

        CASE %ID_CLOSE
          IF hTimer THEN KillTimer hDlg, hTimer  'Kill the timer
          DIALOG END CBHNDL, (CBCTL = %IDOK)     'and end it all

        CASE 100 TO 100 + n 'The bricks - label controls
          LOCAL bc AS LONG: bc = CBCTL - 100: DoPaint = %FALSE

          ' Check to see if the Brick has an empty space next to it
          IF (bPos(bc).z  + 1 = EmptyPos AND (bPos(bc).z + 1) MOD brc <> 0) OR _
             (bPos(bc).z  - 1 = EmptyPos AND (EmptyPos + 1) MOD brc <> 0) OR _
              bPos(bc).z  + brc = EmptyPos OR bPos(bc).z  - brc = EmptyPos THEN

             ' Okay, so move the Brick into the empty space
             CONTROL SET LOC hDlg, CBCTL, bPos(EmptyPos).x, bPos(EmptyPos).y

             IF EmptyPos = bc  THEN  'Was it a "good" position?
               DoPaint = %TRUE: BEEP 'Could play something else here too..
             END IF

             LOCAL tmp AS LONG       'Store the new, empty space
             tmp = EmptyPos: EmptyPos = bPos(bc).z
             bPos(bc).z = tmp

            ' InvalidateRect triggers %WM_CTLCOLORSTATIC
            InvalidateRect CBLPARAM, BYVAL %NULL, 1: UpdateWindow CBLPARAM

            bc = CheckDone
            IF bc >= n THEN          'All is done, build and show a message
              rStr = "Done in "
              rStr = rStr & FORMAT$((TIMER - tTimer)  \  60, "0") & " min. " _
                          & FORMAT$(CLNG(TIMER - tTimer) MOD 60, "0") & " sec."
              rStr = rStr & CHR$(10, 10)
              rStr = rStr & "Play again?" & CHR$(0)

              IF MessageBox(hDlg, BYVAL STRPTR(rStr),"Game" & FORMAT$(n) _
                   & CHR$(0), %MB_YESNO OR %MB_ICONWARNING) = %IDYES THEN
                 CALL NewGame
              ELSE                                     'End game
                IF hTimer THEN KillTimer hDlg, hTimer  'Kill the timer
                DIALOG END CBHNDL, (CBCTL = %IDOK)     'and end it all
              END IF
            END IF
          END IF
      END SELECT

    CASE %WM_CTLCOLORSTATIC 'Set textcolor (Eqv. to VB's "Label.ForeColor")
      IF ISTRUE(DoPaint) THEN
        SetBkMode CBWPARAM, %TRANSPARENT
        SetTextColor CBWPARAM, %BLUE
        FUNCTION = GetStockObject(%LTGRAY_BRUSH)
      ELSE
        SetBkMode CBWPARAM, %TRANSPARENT
        SetTextColor CBWPARAM, %BLACK
        FUNCTION = GetStockObject(%LTGRAY_BRUSH)
      END IF

    CASE %WM_CTLCOLORDLG
      FUNCTION = hBrushDialog   ' Paint the dialog's background
                                ' (Eqv. to VB's "Me.BackColor")

    CASE %WM_SYSCOMMAND
      IF CBWPARAM = %SC_CLOSE THEN
         IF EndProcedure() = 0 THEN
           FUNCTION = 1: EXIT FUNCTION
         END IF
     END IF

    CASE %WM_DESTROY 'Delete the brushes
      DeleteObject hBrushBlack
      DeleteObject hBrushBlue
      DeleteObject hBrushDialog
  END SELECT
END FUNCTION

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' NewGame generates a new game.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
SUB NewGame    'We get smother action with redraw turned off
  CALL SendMessage(hDlg, %WM_SETREDRAW, %FALSE, 0)
  CALL ReMix   'Set new random order
  CALL SendMessage(hDlg, %WM_SETREDRAW, %TRUE, 0)
  InvalidateRect hDlg, BYVAL %NULL, 0: UpdateWindow hDlg
  CALL CheckDone
END SUB

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' ReMix generates a random number table and moves the bricks
' to their new positions.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
SUB ReMix
    LOCAL aa AS LONG, bb AS LONG

    RANDOMIZE                 ' Seed the random number generator
    bPos(0).z = RND(1, n) - 1 ' Generate the first number
    EmptyPos = n              ' Down in the corner..
 
    FOR aa = 1 TO n - 1       ' Make sure all numbers are unique
        bPos(aa).z = RND(1, n) - 1
        FOR bb = 0 TO aa - 1
            IF bPos(bb).z = bPos(aa).z THEN
                bPos(aa).z = RND(1, n) - 1
                bb = -1
            END IF
        NEXT bb
    NEXT aa

   FOR aa = 0 TO n - 1 ' Re-position the bricks
     CONTROL SET LOC hDlg, 100 + aa, bPos(bPos(aa).z).x, bPos(bPos(aa).z).y
   NEXT aa

   tTimer = TIMER      ' Reset the timing of the game
   CONTROL SET TEXT hDlg, %ID_LABEL3, "0 min. 0 sec."
END SUB

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' CheckDone loops through all bricks to see which ones are
' placed correctly and then shows the result in %ID_LABEL1
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
FUNCTION CheckDone AS LONG
   LOCAL aa AS LONG, isDone AS LONG, chDc AS LONG, cc AS LONG

   FOR aa = 0 TO n - 1
     IF aa = bPos(aa).z THEN
       DoPaint = %TRUE       ' Ok, so Blue text color
       INCR isDone
     ELSE
       DoPaint = %FALSE      ' Not ok, use Black text
     END IF

     IF (ISTRUE(DoPaint) AND bPos(aa).c = %BLACK) OR _
        (ISFALSE(DoPaint) AND bPos(aa).c = %BLUE) THEN

       'SET TEXT triggers %WM_CTLCOLORSTATIC, where the color is set
       CONTROL SET TEXT hDlg, 100 + aa, FORMAT$(aa + 1)
     END IF
   NEXT aa

   IF DoPaint THEN            'Set used color to the array
     bPos(aa).c = %BLUE
   ELSE
     bPos(aa).c = %BLACK
   END IF

   DoPaint = %FALSE          ' False, so next color becomes Black
   CONTROL SET TEXT hDlg, %ID_LABEL1, FORMAT$(isDone) & " brick(s) done."
   FUNCTION = isDone         ' Return the result
END FUNCTION

FUNCTION EndProcedure() AS LONG
  IF MessageBox(hDlg, "Exit game?", "Game" & FORMAT$(n) _
     & CHR$(0), %MB_YESNO OR %MB_ICONQUESTION) = %IDYES THEN
     FUNCTION = 1 : EXIT FUNCTION
  END IF
END FUNCTION
